import numpy as np
import pandas as pd


def _safe_hash(df):
    hashed_columns = {}
    for column in df.columns:
        try:
            hashed_columns[column] = pd.util.hash_pandas_object(df[column])
        except TypeError:
            hashed_columns[column] = pd.util.hash_pandas_object(df[column].apply(str))
    return pd.util.hash_pandas_object(pd.concat(hashed_columns, axis=1))


def homomorphic_hash(pandas_object):
    try:
        if isinstance(pandas_object, pd.DataFrame):
            return np.bitwise_xor.reduce(_safe_hash(pandas_object).values)
        return np.bitwise_xor.reduce(pd.util.hash_pandas_object(pandas_object).values)
    except TypeError:
        return hash(pandas_object)



from collections import defaultdict
from functools import wraps
from inspect import isasyncgenfunction, iscoroutinefunction, isgeneratorfunction
from typing import Dict, Type

from uhura.base import Readable, Writable
from uhura.caches import CachingService
from uhura.comparison import BasicComparer
from uhura.composition import async_unit


def cache_local_input(cache_service: CachingService):
    def decorator(readable_cls: Type[Readable]):
        _original_read = readable_cls.read

        @wraps(readable_cls.read)
        def read(self):
            if isasyncgenfunction(_original_read) or isgeneratorfunction(_original_read):
                self._read_count = 0
            cache = cache_service.cache_for_instance(self)
            if cache.exists():
                return cache_service.pull_from_cache(self, _original_read)
            return cache_service.store_and_return(self, _original_read)

        readable_cls.read = read
        return readable_cls

    return decorator


def replace_local_input(cache_service: CachingService):
    def decorator(cls: Type[Readable]):
        _original_read = cls.read

        @wraps(cls.read)
        def read(self):
            if isasyncgenfunction(_original_read) or isgeneratorfunction(_original_read):
                self._read_count = 0
            return cache_service.pull_from_cache(self, _original_read)

        cls.read = read
        return cls

    return decorator


def save_known_good_output(cache_service: CachingService):
    def decorator(writable_cls: Type[Writable]):
        def write(self, obj):
            cache = cache_service.cache_for_instance(self)
            if not cache.exists():
                cache.update(obj)
            pass

        if iscoroutinefunction(writable_cls.write):
            write = async_unit(write)

        writable_cls.write = wraps(writable_cls.write)(write)
        return writable_cls

    return decorator


def compare_known_good_output(
    cache_service: CachingService,
    comparison: Dict = defaultdict(BasicComparer),
):
    def decorator(writable_cls: Type[Writable]):
        def write(self, obj):
            cache = cache_service.cache_for_instance(self)
            if not cache.exists():
                raise FileNotFoundError(f"No fixture found for {writable_cls}.")
            expected = cache.get()
            return comparison[writable_cls.__name__](expected, obj)

        if iscoroutinefunction(writable_cls.write):
            write = async_unit(write)

        writable_cls.write = wraps(writable_cls.write)(write)
        return writable_cls

    return decorator



from contextlib import contextmanager


def get_original(method):
    return getattr(method, "__wrapped__", method)


def is_wrapper(attr):
    return hasattr(attr, "__wrapped__")


class Decorable:
    _decorators = []

    @classmethod
    def _apply_to_current_subclasses(cls, decorator):
        for subclass in cls.__subclasses__():
            decorator(subclass)

    def __init_subclass__(cls, **kwargs):
        cls._decorators = cls._decorators.copy()
        for decorator in cls._decorators:
            decorator(cls)
        super().__init_subclass__(**kwargs)
        pass

    @classmethod
    def add_decorator(cls, decorator):
        cls._apply_to_current_subclasses(decorator)
        cls._decorators.append(decorator)

    @classmethod
    def remove_last_decorator(cls):
        def _revert_decoration(child_cls):
            for key, attr in vars(child_cls).items():
                if is_wrapper(attr) and hasattr(cls, key):
                    setattr(child_cls, key, get_original(attr))

        cls._apply_to_current_subclasses(_revert_decoration)
        cls._decorators.pop()

    @classmethod
    @contextmanager
    def decorated_with(cls, decorator):
        try:
            cls.add_decorator(decorator)
            yield
        finally:
            cls.remove_last_decorator()



import abc
import logging
import os
from abc import abstractmethod
from collections import defaultdict
from inspect import isasyncgenfunction, iscoroutinefunction, isgeneratorfunction
from typing import Callable, DefaultDict, Optional, Type  # ParamSpec

from uhura.composition import async_unit, compose
from uhura.serde import DEFAULT_SERDE, Serde
from uhura.base import Cacheable

logger = logging.getLogger("uhura.caches")


class Cache(abc.ABC):
    @abstractmethod
    def exists(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def get(self):
        raise NotImplementedError()

    @abstractmethod
    def update(self, obj):
        raise NotImplementedError()


class LocalCache(Cache):
    def __init__(self, path: str, serde=DEFAULT_SERDE) -> None:
        self._path = path
        self._serde = serde

    def exists(self):
        return os.path.exists(self._path)

    def update(self, obj):
        logger.info(f"Dumping to {self._path}")

        if not os.path.exists(os.path.dirname(self._path)):
            os.makedirs(os.path.dirname(self._path), exist_ok=True)
        self._serde.write_to_file(self._path, obj)
        return obj

    def get(self):
        return self._serde.read_from_file(self._path)

    @classmethod
    def render_path(cls, base_path: str, cache_key: str, serde: Serde):
        if not cache_key.endswith(serde.file_extension):
            cache_key = cache_key + serde.file_extension
        return os.path.join(base_path, cache_key)

    @classmethod
    def cache_for_instance(cls, cacheable: Cacheable, base_path: str):
        cache_key = cacheable.cache_key()
        serde = cacheable.get_serde()
        path = cls.render_path(base_path, cache_key, serde)
        return cls(path, serde)


async def _take(async_iterable, n=5):
    index = 0
    async for item in async_iterable:
        index += 1
        if index > n:
            break
        yield item


def _sync_take(iterable, n=5):
    index = 0
    for item in iterable:
        index += 1
        if index > n:
            break
        yield item


class CachingService:
    def __init__(self, lookup, base_path, depth=4):
        self._base_path = base_path
        self._depth = depth
        self._lookup: DefaultDict[str, Callable[[Cacheable, str, ...], Cache]] = lookup

    def _iterate_from_cache(self, instance: Cacheable):
        for _ in range(self._depth):
            cache = self.cache_for_instance(instance)
            yield cache.get()
            instance._read_count += 1

    def cache_for_instance(self, instance: Cacheable):
        return self._lookup[instance.cache_key()](instance, self._base_path)

    def store_and_return(self, instance: Cacheable, method):
        def _update(value):
            cache = self.cache_for_instance(instance)
            if cache.exists():
                return cache.get()
            else:
                cache.update(value)
                return value

        if isasyncgenfunction(method):
            return _take(compose(method, _update)(instance), n=self._depth)
        if isgeneratorfunction(method):
            return _sync_take(compose(method, _update)(instance), n=self._depth)
        return compose(method, _update)(instance)

    def pull_from_cache(self, instance: Cacheable, original_method):
        if isasyncgenfunction(original_method):
            return async_unit(self._iterate_from_cache(instance))
        if isgeneratorfunction(original_method):
            return self._iterate_from_cache(instance)
        if iscoroutinefunction(original_method):
            return async_unit(self.cache_for_instance(instance).get())
        return self.cache_for_instance(instance).get()

    @classmethod
    def default(cls, base_path, **kwargs):
        return cls(
            lookup=defaultdict(lambda: LocalCache.cache_for_instance),
            base_path=base_path,
            **kwargs,
        )



import asyncio
import os
from abc import ABC, abstractmethod
from inspect import isasyncgenfunction, iscoroutinefunction, isgeneratorfunction, unwrap
from typing import Generic, Type, TypeVar, Optional

from uhura.serde import DEFAULT_SERDE, Serde
from uhura.decorables import Decorable

ReadType = TypeVar("ReadType")
WriteType = TypeVar("WriteType")


class Cacheable(ABC):
    _read_count: Optional[int]  # Created in streaming readables (generators, async iterators etc.)

    @abstractmethod
    def cache_key(self):
        if hasattr(self, "_read_count"):
            return os.path.join(self.__class__.__name__, str(self._read_count))
        return self.__class__.__name__

    @abstractmethod
    def get_serde(self) -> Serde:
        raise NotImplementedError()


class Readable(Decorable, Cacheable, Generic[ReadType]):
    def __init_subclass__(cls) -> None:
        if isasyncgenfunction(cls.read):
            cls.read = _as_stream_read(cls.read)
        elif isgeneratorfunction(cls.read):
            cls.read = _as_sync_stream_read(cls.read)
        super().__init_subclass__()

    @abstractmethod
    async def read(self) -> ReadType:
        raise NotImplementedError()

    def sync_read(self) -> ReadType:
        if iscoroutinefunction(unwrap(self.read)):
            return asyncio.run(self.read())
        return self.read()

    def get_serde(self):
        return DEFAULT_SERDE

    def cache_key(self):
        if hasattr(self, "_read_count"):
            return f"{os.path.join(self.__class__.__name__, str(self._read_count))}{self.get_serde().file_extension}"
        return f"{self.__class__.__name__}{self.get_serde().file_extension}"


def _as_stream_read(read_method):
    async def read(self):
        self._read_count = 0
        async for item in read_method(self):
            yield item
            self._read_count += 1

    return read


def _as_sync_stream_read(read_method):
    def read(self):
        self._read_count = 0
        for item in read_method(self):
            yield item
            self._read_count += 1

    return read


class Writable(Decorable, Cacheable, Generic[WriteType]):
    @abstractmethod
    async def write(self, obj: WriteType):
        raise NotImplementedError()

    def sync_write(self, obj: WriteType):
        if iscoroutinefunction(unwrap(self.write)):
            return asyncio.run(self.write(obj))
        return self.write(obj)

    def get_serde(self):
        return DEFAULT_SERDE

    def cache_key(self):
        return f"{self.__class__.__name__}{self.get_serde().file_extension}"



import pytest
from functools import wraps

from uhura.properties.transformer import transformer
from uhura.properties.property_tester import AbstractPropertyTester


class _DummyTester(AbstractPropertyTester):
    def __init__(self):
        self.intercepted_values = []

    def add_properties_to_function(self, tested):
        @wraps(tested)
        def wrapper(arg):
            self.intercepted_values.append(arg)
            return tested(arg)

        return wrapper


def test_simple_transformer():
    @transformer
    def modify(val: int):
        return val + 2

    intercepted_values = []

    assert modify(10) == 12
    assert intercepted_values == []
    modify.tester = _DummyTester()
    assert modify(12) == 14
    assert modify.tester.intercepted_values == [12]


def test_raises_if_no_argname():
    with pytest.raises(AssertionError):

        @transformer
        def add(a: int, b: int):
            return a + b


def test_raises_if_arg_name_not_in_sig():
    with pytest.raises(AssertionError):

        @transformer(evaluated_arg_name="hello")
        def add(a: int, b: int):
            return a + b


def test_single_arg():
    @transformer(evaluated_arg_name="b")
    def add(a: int, b: int):
        return a + b

    intercepted_values = []
    assert add(1, 2) == 3
    assert intercepted_values == []
    add.tester = _DummyTester()
    assert add(4, 5) == 9
    assert add.tester.intercepted_values == [5]



import pytest
import random
import numpy as np
import string
import pandas as pd


@pytest.fixture
def base_df():
    df_len = 20
    return pd.DataFrame(
        {
            "a": np.random.randn(df_len),
            "b": random.choices(string.ascii_letters, k=df_len),
            "c": np.repeat(np.nan, df_len),
            "d": range(df_len),
            "b_idx": random.choices(string.ascii_letters, k=df_len),
            "d_idx": range(df_len),
        }
    ).set_index(["b_idx", "d_idx"])



from contextlib import ExitStack, contextmanager

from uhura.base import Readable, Writable
from uhura.caches import (
    CachingService,
)
from uhura.caching import (
    cache_local_input,
    compare_known_good_output,
    replace_local_input,
    save_known_good_output,
)
from uhura.properties.property_tester import PropertyTester
from uhura.properties.transformer import Transformer

try:
    from uhura.pandas_tools import pandas_comparator as default_comparator
    from uhura.properties.testable_properties import DEFAULT_PROPERTIES
except ImportError:
    from uhura.caches import basic_comparator as default_comparator

    DEFAULT_PROPERTIES = tuple()


@contextmanager
def _compose_contexts(*contexts):
    with ExitStack() as stack:
        for context in contexts:
            stack.enter_context(context)
        yield


def fixture_builder_mode(
    *,
    known_good_path: str = "tests/fixtures/output_known_good",
    input_path: str = "tests/fixtures/input",
    stream_cache_depth: int = 3,
):
    return _compose_contexts(
        Readable.decorated_with(
            cache_local_input(
                CachingService.default(base_path=input_path, depth=stream_cache_depth)
            )
        ),
        Writable.decorated_with(
            save_known_good_output(CachingService.default(base_path=known_good_path))
        ),
    )


def task_test_mode(
    *,
    known_good_path: str = "tests/fixtures/output_known_good",
    input_path: str = "tests/fixtures/input",
    comparison=default_comparator,
):
    return _compose_contexts(
        Readable.decorated_with(replace_local_input(CachingService.default(base_path=input_path))),
        Writable.decorated_with(
            compare_known_good_output(
                CachingService.default(base_path=known_good_path), comparison=comparison
            )
        ),
    )


_tester = PropertyTester.create_for_properties(DEFAULT_PROPERTIES)


@contextmanager
def test_transformers(tester=_tester):
    """
    The test_transformers mode activates property testing for all functions annotated as 'transformers'.

    To use specific tests, or a specific testing mechanism, use the optional 'tester' argument to provide an
    alternative.
    """
    Transformer.tester = tester
    try:
        yield
    finally:
        Transformer.tester = None



import os
import tempfile

from uhura.functional import uhura_reader, uhura_writer
from uhura.modes import fixture_builder_mode


def test_readers_attach_pickle_extension():
    @uhura_reader
    def readable_thing_1():
        return "hello"

    with tempfile.TemporaryDirectory() as temp_dir:
        output_path = os.path.join(temp_dir, "good_output")
        input_path = os.path.join(temp_dir, "input")

        with fixture_builder_mode(known_good_path=output_path, input_path=input_path):
            readable_thing_1()
        assert os.path.exists(os.path.join(input_path, f"{readable_thing_1.__qualname__}.pkl"))


def test_writers_attach_pickle_extension():
    @uhura_writer
    def writable_thing_1(thing: str):
        pass

    with tempfile.TemporaryDirectory() as temp_dir:
        output_path = os.path.join(temp_dir, "good_output")
        input_path = os.path.join(temp_dir, "input")

        with fixture_builder_mode(known_good_path=output_path, input_path=input_path):
            writable_thing_1("goodbye")
        assert os.path.exists(os.path.join(output_path, f"{writable_thing_1.__qualname__}.pkl"))



from uhura.comparison import BasicComparer
from operator import methodcaller
import pytest


def test_basic_comparisons():
    comparer = BasicComparer()
    assert comparer(1, 1)
    assert comparer("hello", "hello")
    with pytest.raises(AssertionError):
        comparer(1, "1")
    with pytest.raises(AssertionError):
        comparer("Hello", "hEllo")


def test_actual_preprocessing():
    comparer = BasicComparer()
    comparer.add_actual_preprocessor(int)
    assert comparer("1", 1)


def test_expected_preprocessing():
    comparer = BasicComparer()
    comparer.add_expected_preprocessor(int)
    assert comparer(1, "1")


def test_preprocessing():
    comparer = BasicComparer()
    comparer.add_preprocessor(methodcaller("upper"))
    assert comparer("Hello", "hEllo")



import logging

import pandas as pd
import numpy as np
import functools
import pytest
from typing import Optional, Mapping
import string
import random

from uhura.properties import testable_properties as properties
from uhura.properties import transformer, PropertyTester
from uhura.modes import test_transformers


@pytest.fixture
def base_df():
    df_len = 20
    return pd.DataFrame(
        {
            "a": np.random.randn(df_len),
            "b": random.choices(string.ascii_letters, k=df_len),
            "c": np.repeat(np.nan, df_len),
            "d": range(df_len),
            "b_idx": random.choices(string.ascii_letters, k=df_len),
            "d_idx": range(df_len),
        }
    ).set_index(["b_idx", "d_idx"])


def sequential_counter(df: pd.DataFrame) -> pd.DataFrame:
    df = df.copy(deep=True)
    df["counter"] = range(0, df.shape[0])
    return df


def mapper(df: pd.DataFrame, mapper: Optional[Mapping] = None) -> pd.DataFrame:
    mapper = mapper if mapper else {l: l.upper() for l in string.ascii_letters}
    df = df.copy(deep=True)
    df["c"] = df["b"].map(mapper)
    return df


def drop_col(df: pd.DataFrame, col: str) -> pd.DataFrame:
    df = df.copy(deep=True)
    return df.drop(col, axis=1)


def drop_rows(df: pd.DataFrame) -> pd.DataFrame:
    df = df.copy(deep=True)
    return df.iloc[:3]


def convert_to_int(df: pd.DataFrame) -> pd.DataFrame:
    df = df.copy(deep=True)
    df["a"] = df["a"].astype(int)
    return df


def convert_index_to_int(df: pd.DataFrame) -> pd.DataFrame:
    df = df.copy(deep=True)
    idx = df.index
    df = df.reset_index()
    df[idx.names[-1]] = df[idx.names[-1]].astype(str)
    return df.set_index(idx.names)


def add_random(df: pd.DataFrame) -> pd.DataFrame:
    df = df.copy(deep=True)
    df["random_column"] = np.random.random(len(df))
    return df


def inplace_add_random(df: pd.DataFrame) -> pd.DataFrame:
    df["random_column"] = np.random.random(len(df))
    return df


logging.getLogger().setLevel(logging.INFO)


@pytest.mark.parametrize(
    "property_test,transformer,expected",
    [
        (properties.homomorphic, mapper, True),
        (properties.homomorphic, sequential_counter, False),
        (properties.idempotent, mapper, True),
        (properties.idempotent, add_random, False),
        (properties.order_invariant, mapper, True),
        (properties.order_invariant, sequential_counter, False),
        (properties.adds_columns, sequential_counter, True),
        (properties.drops_columns, functools.partial(drop_col, col="a"), True),
        (properties.preserves_row_count, sequential_counter, True),
        (properties.preserves_row_count, drop_rows, False),
        (properties.preserves_types, drop_rows, True),
        (properties.preserves_types, convert_to_int, False),
        (properties.preserves_index_types, drop_rows, True),
        (properties.preserves_index_types, convert_index_to_int, False),
        (properties.pure, add_random, True),
        (properties.pure, inplace_add_random, False),
    ],
)
def test_decorators(property_test, transformer, expected, base_df):
    result = transformer(base_df) if property_test is not properties.pure else None
    _, observed = property_test(transformer, arg=base_df, result=result)
    assert expected == observed


def test_columns_not_modified(base_df):
    base_df_copy = base_df.copy(deep=True)
    columns_not_modified = properties.get_columns_not_modified_property(
        ["a", "c", "d", "missing_col"]
    )

    @transformer()
    def modify_and_drop(_df):
        _df = _df[[c for c in _df.columns if c != "c"]]
        _df.loc[_df["d"] == 1, "a"] = -10
        modify_and_drop.counter += 1
        return _df

    modify_and_drop.counter = 0

    @transformer()
    def modfify_other_columns(_df):
        _df["b"] = 10
        return _df

    _tester = PropertyTester.create_for_properties([columns_not_modified])
    with test_transformers(tester=_tester):
        modify_and_drop(base_df.copy(deep=True))
        # test we call twice to test order invariance and that we never call for testing columns modification
        assert modify_and_drop.counter == 2
        results = _tester.get_results(modify_and_drop, summary=False)
        assert results["order_invariant"].has_property
        assert (
            results["get_columns_not_modified_property.<locals>.columns_not_modified"].has_property
            is False
        )
        assert (
            results["get_columns_not_modified_property.<locals>.columns_not_modified"].notes
            == "a modified, c dropped"
        )

        modfify_other_columns(base_df)
        # test that modified column not being tracked is being ignored
        results = _tester.get_results(modfify_other_columns, summary=False)
        assert results["order_invariant"].has_property
        assert results[
            "get_columns_not_modified_property.<locals>.columns_not_modified"
        ].has_property
        assert (
            results["get_columns_not_modified_property.<locals>.columns_not_modified"].notes == ""
        )
        # test that the original index sorting is being preserved even if order invariance is being tested
        pd.testing.assert_index_equal(base_df_copy.index, base_df.index)


def test_property_only_checked_once_with_nested_functions(base_df, caplog):
    """As described in the `_Transformer.__call__` by checking properties
    that require calling the same function several times we may lead to
    combinatorial explosion. This test makes sure that we have a mechanism
    for handling it.
    As of the current implementation homorphism will run the function it
    decorates 6 times. In the case of this test the function being called
    is `function_with_log`. `function_with_log` then calls mapper.
    So, `mapper` will already be called 6 times due to the 6
    `function_with_log` calls. `mapper` then is also checking homomorphism
    on itself. So, the property is going to do 6 `mapper` calls each time
    is being called. So, we would have 6 `function_with_log` that will
    trigger 6 `mapper` calls each, which will lead to 36 mapper calls overall.
    Effectively we only need to test the properties for `mapper` once
    and then call `mapper` only once per `function_with_log` call.
    This should lead to 6 `mapper` calls when its property is checked
    the first time and then 5 other `mapper` calls, one per
    `function_with_log` we need to do to test its homomorphic property

    Also, we are going to call the `mapper` functions another 2 times when
    testing that `function_with_log` and `mapper` and pure. As well as
    other 2 times when running the actual functions in `wrapper`

    """

    @transformer(evaluated_arg_name="df")
    def mapper(df: pd.DataFrame, mapping_dict: Optional[Mapping] = None) -> pd.DataFrame:
        mapping_dict = (
            mapping_dict if mapping_dict else {l: l.upper() for l in string.ascii_letters}
        )
        df = df.copy(deep=True)
        df["c"] = df["b"].map(mapping_dict)
        logging.info("mapper logging")
        mapper.counter += 1
        return df

    mapper.counter = 0

    @transformer(evaluated_arg_name="df")
    def function_with_log(df: pd.DataFrame) -> pd.DataFrame:
        logging.info("function_with_log logging")
        return mapper(df)

    _tester = PropertyTester.create_for_properties([properties.homomorphic])

    with test_transformers(tester=_tester):
        function_with_log(base_df)
        assert mapper.counter == 11
        logging.info("original logging levels restored")
        expected_logs = [
            "function_with_log logging",
            "mapper logging",
            "original logging levels restored",
        ]
        assert len(caplog.messages) == 7
        assert caplog.messages[:2] == expected_logs[:2]
        assert caplog.messages[-1:] == expected_logs[-1:]
        results = _tester.get_results(function_with_log, summary=False)
        assert results["pure"].has_property
        assert results["homomorphic"].has_property

        results = _tester.get_results(mapper, summary=False)
        assert results["pure"].has_property
        assert results["homomorphic"].has_property


def _identity(df):
    return df


def test_preserves_index_types_for_simple_index():
    df = pd.DataFrame(data={"a": list(range(10)), "b": list(range(10))})
    result = _identity(df)
    assert properties.preserves_index_types(_identity, df, result=result)[1]


def test_custom_debugger():
    def how_has_index_changed(arg, result=None):
        if result is not None:
            missing_in_arg = set(result.index) - set(arg.index)
            missing_in_result = set(arg.index) - set(result.index)

            return f"indexes missing in arg: {list(missing_in_arg)} indexes missing in result {list(missing_in_result)}"
        else:
            return f"arg shape {arg.shape}"

    custom_debugger = properties.create_custom_debugger(how_has_index_changed)

    @transformer
    def rename_indexes(df):
        df.rename(index={8: 10, 9: 11}, inplace=True)
        return df

    base_df = pd.DataFrame({"a": range(10)})
    _tester = PropertyTester.create_for_properties([custom_debugger])

    with test_transformers(tester=_tester):
        rename_indexes(base_df)
        results = _tester.get_results(rename_indexes, summary=False)
        assert (
            results["create_custom_debugger.<locals>.debugger"].notes
            == " pre [rename_indexes]: - arg shape (10, 1) post [rename_indexes]: - indexes missing in arg: [10, 11] indexes missing in result [8, 9]"
        )



import asyncio
import operator
import os
import random
import tempfile
from inspect import isasyncgen
from typing import AsyncIterable, Iterable

import pytest

from uhura.base import Readable, Writable
from uhura.caches import (
    CachingService,
    LocalCache,
)
from uhura.caching import (
    cache_local_input,
    compare_known_good_output,
    replace_local_input,
    save_known_good_output,
)
from uhura.functional import uhura_reader, uhura_writer
from uhura.modes import fixture_builder_mode, task_test_mode
from uhura.serde import PickleSerde


class RandomReadable(Readable[float]):
    async def read(self) -> float:
        return random.random()


@uhura_reader
async def random_function():
    return random.random()


@uhura_reader
def sync_random_function():
    return random.random()


class SyncRandomReadable(Readable[float]):
    def read(self) -> float:
        return random.random()


@uhura_reader
def sync_random_stream():
    for _ in range(10):
        yield random.random()


@uhura_reader
async def random_stream():
    for _ in range(10):
        yield random.random()


class RandomStreamReadable(Readable[float]):
    async def read(self) -> AsyncIterable[float]:
        for _ in range(10):
            yield random.random()


class SyncRandomStreamReadable(Readable[float]):
    def read(self) -> Iterable[float]:
        for _ in range(10):
            yield random.random()


class DummyWritable(Writable[int]):
    async def write(self, obj):
        pass


@uhura_writer
async def dummy_write(obj):
    pass


@uhura_writer
def sync_dummy_write(obj):
    pass


def dummy_write_with_arg(arg_value="test"):
    @uhura_writer(output_arg="obj")
    def dummy_write(obj=None, arg=arg_value):
        pass

    return dummy_write


class SyncDummyWritable(Writable[int]):
    def write(self, obj):
        pass


def _as_sync(coro):
    def inner(*args, **kwargs):
        return asyncio.run(coro(*args, **kwargs))

    return inner


@pytest.mark.parametrize(
    "read_method",
    [
        RandomReadable().sync_read,
        SyncRandomReadable().sync_read,
        _as_sync(random_function),
        sync_random_function,
    ],
)
def test_simple_input_caching(read_method):
    uncached_read = read_method()
    with tempfile.TemporaryDirectory() as temp_dir:
        cache_service = CachingService.default(base_path=temp_dir)
        with Readable.decorated_with(cache_local_input(cache_service)):
            first_cached_read = read_method()
            assert (
                uncached_read != first_cached_read
            ), "Random readable is always returning the same thing"
            second_cached_read = read_method()
            assert first_cached_read == second_cached_read, "Readable not being cached"
        with Readable.decorated_with(replace_local_input(cache_service)):
            read_from_previous_cache = read_method()
            assert (
                first_cached_read == read_from_previous_cache
            ), "Reading from simple cache failed to return the same result"
    post_cache_read = read_method()
    assert post_cache_read != second_cached_read, "Removing simple cache failed"
    pass


async def _take_5(stream_read):
    items = []
    async for item in stream_read:
        items.append(item)
        if len(items) == 5:
            break
    return items


def _get_5_from_stream(stream):
    if isasyncgen(stream):
        return asyncio.run(_take_5(stream))
    items = []
    for item in stream:
        items.append(item)
        if len(items) == 5:
            break
    return items


def _make_look_like_a_normal_readable(function):
    function.read = function
    return function


@pytest.mark.parametrize(
    "stream_readable",
    [
        RandomStreamReadable(),
        SyncRandomStreamReadable(),
        _make_look_like_a_normal_readable(sync_random_stream),
        _make_look_like_a_normal_readable(random_stream),
    ],
)
def test_stream_input_caching(stream_readable):
    uncached_read = _get_5_from_stream(stream_readable.read())
    with tempfile.TemporaryDirectory() as temp_dir:
        cache_service = CachingService.default(base_path=temp_dir, depth=5)
        with Readable.decorated_with(cache_local_input(cache_service)):
            first_cache_read = _get_5_from_stream(stream_readable.read())
            assert sorted(uncached_read) != sorted(
                first_cache_read
            ), "Random readable is always returning the same thing"
            second_cache_read = _get_5_from_stream(stream_readable.read())
            assert sorted(first_cache_read) == sorted(
                second_cache_read
            ), "Stream Readable is not being cached"
        with Readable.decorated_with(replace_local_input(cache_service)):
            read_from_previous_cache = _get_5_from_stream(stream_readable.read())
            assert sorted(first_cache_read) == sorted(
                read_from_previous_cache
            ), "Reading from cache failed to return the same result"
    post_cache_read = _get_5_from_stream(stream_readable.read())
    assert sorted(post_cache_read) != sorted(read_from_previous_cache), "Removing cache failed"
    pass


@pytest.mark.parametrize(
    "write_method",
    [
        DummyWritable().sync_write,
        SyncDummyWritable().sync_write,
        sync_dummy_write,
        dummy_write_with_arg("test"),
        _as_sync(dummy_write),
    ],
)
def test_output_caching_and_comparison(write_method):
    with tempfile.TemporaryDirectory() as temp_dir:
        cache_service = CachingService.default(base_path=temp_dir)
        with Writable.decorated_with(save_known_good_output(cache_service)):
            write_method(1)
        with Writable.decorated_with(compare_known_good_output(cache_service)):
            with pytest.raises(AssertionError):
                write_method(2)  # Test raises if different
            write_method(1)  # But not if the same
    write_method(123)  # Test returns to normal write afterwards


@pytest.mark.parametrize(
    ["simple_read", "stream_read", "writer"],
    [
        (RandomReadable(), RandomStreamReadable(), DummyWritable()),
        (SyncRandomReadable(), SyncRandomStreamReadable(), SyncDummyWritable()),
    ],
)
def test_missing_files(simple_read, stream_read, writer):
    with tempfile.TemporaryDirectory() as temp_dir:
        cache_service = CachingService.default(base_path=temp_dir)
        with Readable.decorated_with(cache_local_input(cache_service)):
            with Writable.decorated_with(save_known_good_output(cache_service)):
                simple_read.sync_read()
                _get_5_from_stream(stream_read.read())
                writer.sync_write(1)

        Readable.add_decorator(replace_local_input(cache_service))
        Writable.add_decorator(compare_known_good_output(cache_service))

    with pytest.raises(FileNotFoundError):
        simple_read.sync_read()

    with pytest.raises(FileNotFoundError):
        _get_5_from_stream(stream_read.read())

    with pytest.raises(FileNotFoundError):
        writer.sync_write(1)

    Writable.remove_last_decorator()
    Readable.remove_last_decorator()


@pytest.mark.parametrize(
    ["simple_read", "stream_read", "writer"],
    [
        (RandomReadable(), RandomStreamReadable(), DummyWritable()),
        (SyncRandomReadable(), SyncRandomStreamReadable(), SyncDummyWritable()),
    ],
)
def test_modes(simple_read, stream_read, writer):
    _initial_strategy = operator.add
    _changed_strategy = operator.mul

    def _important_function(strategy=_initial_strategy):
        addition = simple_read.sync_read()
        series = _get_5_from_stream(stream_read.read())
        writer.sync_write(strategy(series[0], addition))
        pass

    with tempfile.TemporaryDirectory() as temp_dir:
        output_path = os.path.join(temp_dir, "good_output")
        input_path = os.path.join(temp_dir, "input")

        with fixture_builder_mode(known_good_path=output_path, input_path=input_path):
            _important_function()

        with task_test_mode(known_good_path=output_path, input_path=input_path):
            _important_function()
            with pytest.raises(AssertionError):
                _important_function(_changed_strategy)  # You've broken something!


_fake_database = {"id1": 27, "id2": 34}


class DatabaseReader(Readable[float]):
    def __init__(self, id_: str):
        self._id = id_

    def read(self):
        return _fake_database[self._id] + random.random()

    def cache_key(self):
        return f"DatabaseReader/{self._id}"


def test_cache_key():
    with fixture_builder_mode():
        assert DatabaseReader("id1").read() == DatabaseReader("id1").read()
        assert DatabaseReader("id2").read() == DatabaseReader("id2").read()
        assert (
            DatabaseReader("id1").read() != DatabaseReader("id2").read()
        ), "Cache key being ignored"


def test_local_cache_path_renders_path_with_file_extension_from_serde_if_not_present():
    cache_path = "my/cache/dir"
    serde = PickleSerde()
    local_cache = LocalCache(cache_path, serde)

    base_path = "base/path"
    cache_key = "filename"
    rendered_path = local_cache.render_path(base_path, cache_key, serde)

    assert rendered_path.split(".")[-1] == "pkl"



from abc import ABC, abstractmethod
from typing import Callable


class Comparer(ABC):
    def __init__(self):
        self._actual_prep = []
        self._expected_prep = []

    def add_actual_preprocessor(self, preprocessor: Callable):
        """Add a preprocessing step to be applied to actual values before they are compared."""
        self._actual_prep.append(preprocessor)

    def add_expected_preprocessor(self, preprocessor: Callable):
        """Add a preprocessing step to be applied to expected values before they are compared."""
        self._expected_prep.append(preprocessor)

    def add_preprocessor(self, preprocessor: Callable):
        """Add a preprocessing step to be applied to both values before they are compared."""
        self.add_actual_preprocessor(preprocessor)
        self.add_expected_preprocessor(preprocessor)

    @abstractmethod
    def base_compare(self, actual, expected):
        raise NotImplementedError()

    def _prep_actual(self, actual):
        for prep in self._actual_prep:
            actual = prep(actual)
        return actual

    def _prep_expected(self, expected):
        for prep in self._expected_prep:
            expected = prep(expected)
        return expected

    def __call__(self, actual, expected):
        return self.base_compare(self._prep_actual(actual), self._prep_expected(expected))


class BasicComparer(Comparer):
    def base_compare(self, actual, expected):
        match = actual == expected
        assert match, f"Expected {expected}, found {actual}"
        return match



import pandas as pd

from uhura.properties.property_tester import PropertyTester, depends
from uhura.modes import test_transformers
from uhura.properties.transformer import transformer
from uhura.properties.testable_properties import pure, idempotent
import pytest
from random import random


def _example_callable_factory(name, has_prop):
    def inner(tested, arg, result):
        return None, has_prop

    inner.__qualname__ = name
    return inner


def _explode_if_called(_, __):
    raise NotImplementedError()


@pytest.fixture
def example_properties():
    part_a = _example_callable_factory("a", True)
    part_b = depends(part_a)(_example_callable_factory("b", False))
    part_c = depends(part_b)(_explode_if_called)
    part_d = depends(part_a)(_example_callable_factory("d", True))
    part_e = depends(part_a)(_example_callable_factory("e", True))
    part_f = depends(part_d, part_e)(_example_callable_factory("f", True))
    part_g = depends(part_d, part_e)(_example_callable_factory("g", False))
    return tuple([part_g, part_f, part_e, part_d, part_c, part_b, part_a])


def test_tester(example_properties):
    tester = PropertyTester.create_for_properties(example_properties)

    def tested(_):
        pass

    wrapped_tested = tester.add_properties_to_function(tested)
    wrapped_tested(None)
    assert tester.get_repository(tested).positive_properties == frozenset(["a", "d", "e", "f"])


def test_null_depends():
    with pytest.raises(ValueError):
        depends()(_example_callable_factory("a", True))


@transformer
def dummy_transformer(arg):
    return bool(arg)


@transformer
def dummy_transformer_2(_):
    return random()


def dummy_property_with_notes(tested, arg: pd.DataFrame, result=None):
    return arg, True, "Love letter from property"


def test_transformer_test_mode():
    dummy_tester = PropertyTester.create_for_properties([pure, idempotent])
    with test_transformers(tester=dummy_tester):
        dummy_transformer("yes")
        dummy_transformer_2(None)
    assert dummy_tester.get_results(dummy_transformer) == frozenset({"pure", "idempotent"})
    assert dummy_tester.get_results(dummy_transformer_2) == frozenset({"pure"})


def test_property_with_notes(base_df, caplog):
    dummy_tester = PropertyTester.create_for_properties([dummy_property_with_notes])
    decorated_tester = dummy_tester.add_properties_to_function(lambda arg: arg)
    decorated_tester(base_df)
    assert caplog.messages == [
        "[test_property_with_notes.<locals>.<lambda>]: has property dummy_property_with_notes - Love letter from property"
    ]



import pickle
from typing import Any, ClassVar, Generic, TypeVar

import pandas as pd

SerdeType = TypeVar("SerdeType")


class Serde(Generic[SerdeType]):
    file_extension: ClassVar[str]

    def __init_subclass__(cls) -> None:
        assert hasattr(
            cls, "file_extension"
        ), "Serde implementations must have a valid file_extension"
        return super().__init_subclass__()

    def read_from_file(self, file) -> SerdeType:
        raise NotImplementedError()

    def write_to_file(self, file, obj: SerdeType):
        raise NotImplementedError()


class PickleSerde(Serde[Any]):
    file_extension = ".pkl"

    def read_from_file(self, file) -> SerdeType:
        with open(file, "rb") as infile:
            return pickle.load(infile)

    def write_to_file(self, file, obj: SerdeType):
        with open(file, "wb") as outfile:
            return pickle.dump(obj, outfile)


DEFAULT_SERDE = PickleSerde()


class ParquetSerde(Serde[pd.DataFrame]):
    file_extension = ".parquet"  # Used in cache key

    def __init__(self, read_kwargs=None, write_kwargs=None):
        self._read_kwargs = read_kwargs if read_kwargs is not None else {}
        self._write_kwargs = write_kwargs if write_kwargs is not None else {}

    def read_from_file(self, file: str) -> pd.DataFrame:
        return pd.read_parquet(file, **self._read_kwargs)

    def write_to_file(self, file: str, obj: pd.DataFrame):
        return obj.to_parquet(file, **self._write_kwargs)



import os
from tempfile import TemporaryDirectory
from typing import Any

import pytest

from uhura.base import Readable
from uhura.modes import fixture_builder_mode, task_test_mode
from uhura.serde import PickleSerde, Serde


class FakeSerde(PickleSerde):
    file_extension = ".blah"

    def __init__(self):
        self.has_been_read = False
        self.has_been_written = False

    def read_from_file(self, file):
        self.has_been_read = True
        return super().read_from_file(file)

    def write_to_file(self, file, obj):
        self.has_been_written = True
        return super().write_to_file(file, obj)


def test_can_have_custom_serde():
    test_serde = FakeSerde()

    class FakeReader(Readable):
        def read(self):
            return 1

        def get_serde(self):
            return test_serde

    with TemporaryDirectory() as root:
        with fixture_builder_mode(
            input_path=os.path.join(root, "input"), known_good_path=os.path.join(root, "output")
        ):
            client = FakeReader()
            assert not test_serde.has_been_read
            assert not test_serde.has_been_written
            client.read()
            assert test_serde.has_been_written
            assert os.path.exists(os.path.join(root, "input", "FakeReader.blah"))
            assert not test_serde.has_been_read
            client.read()
            assert test_serde.has_been_read


def test_serde_subclass_requires_file_extension():
    with pytest.raises(AssertionError):

        class Bad(Serde[Any]):
            def read_from_file(self, file):
                pass

            def write_to_file(self, file, obj: Any):
                pass



import logging
from abc import ABCMeta, abstractmethod
from typing import NamedTuple, Optional, Callable, TypeVar, Tuple, Collection
from collections import defaultdict
from contextlib import contextmanager
from graphlib import TopologicalSorter


ARG = TypeVar("ARG")
PropertyTest = Callable[[Callable[[ARG], ARG], ARG], Tuple[ARG, bool]]
UHURA_PROPERTIES_LOGGER_NAME = "uhura.properties"
logger = logging.getLogger(UHURA_PROPERTIES_LOGGER_NAME)
logger.setLevel(logging.INFO)


SENTINEL = object()  # unique object used to signal we have never ran `tested`


@contextmanager
def set_loggers_to_error_only():
    """Testing properties may lead to call the same function multiple
    times and also call all the other functions the tested one will call.
    Obviously we can to limit the number of unnecessary calls,
    but we won't be able to ensure we do only one call per function. So,
    we want to suppress the logs for all the extra calls we do.

    Also, we want to keep on showing error logs in case we have some
    unexpected issues while testing a property that the property itself
    is not accounting for. This should make it easier to debug properties
    """
    loggers = {
        name: logging.getLogger(name)
        for name in logging.root.manager.loggerDict
        if name != UHURA_PROPERTIES_LOGGER_NAME
    }

    logging_levels = {}
    for name, logger in loggers.items():
        logging_levels[name] = logger.getEffectiveLevel()
        logger.setLevel(logging.ERROR)
    root_logger_level = logging.getLogger().getEffectiveLevel()
    logging.getLogger().setLevel(logging.ERROR)

    try:
        yield logging_levels
    finally:
        for name, level in logging_levels.items():
            logging.getLogger(name).setLevel(level)
            logging.getLogger().setLevel(root_logger_level)


class PropertyTestResult(NamedTuple):
    prop: Callable
    has_property: bool
    tested_callable_name: str
    notes: Optional[str] = None

    @property
    def property_name(self):
        return self.prop.__qualname__

    def write_to_info(self):
        respected = "has property" if self.has_property else "lacks property"
        base_message = f"[{self.tested_callable_name}]: {respected} {self.property_name}"
        logging.getLogger(UHURA_PROPERTIES_LOGGER_NAME).info(
            base_message if self.notes is None else f"{base_message} - {self.notes}"
        )
        pass


class HierarchicalPropertyRepository:
    def __init__(self, property_graph: TopologicalSorter):
        self._property_graph = property_graph
        self._results = dict()
        self._valid_properties = set()

    def add_result(self, result: PropertyTestResult):
        if result.has_property:
            self._valid_properties.add(result.property_name)
        self._property_graph.done(result.prop)
        self._results[result.property_name] = result
        pass

    def get_properties_to_test(self) -> PropertyTest:
        while self._property_graph.is_active():
            for property_ in self._property_graph.get_ready():
                if all(
                    dependency.__qualname__ in self._valid_properties
                    for dependency in getattr(property_, "depends_on", set())
                ):
                    yield property_
                else:
                    self._property_graph.done(property_)

    @property
    def positive_properties(self):
        return frozenset(self._valid_properties)

    @property
    def results(self):
        return self._results

    @classmethod
    def create_for_properties(cls, properties: Collection[PropertyTest]):
        graph = {prop: getattr(prop, "depends_on", set()) for prop in properties}
        sorter = TopologicalSorter(graph)
        sorter.prepare()
        return cls(sorter)


class VerbosePropertyRepository(HierarchicalPropertyRepository):
    def add_result(self, result: PropertyTestResult):
        result.write_to_info()
        super().add_result(result)


class AbstractPropertyTester(metaclass=ABCMeta):
    @abstractmethod
    def add_properties_to_function(self, tested):
        pass


class PropertyTester(AbstractPropertyTester):
    def __init__(self, repository: defaultdict):
        self._repository = repository

    def get_repository(self, tested) -> HierarchicalPropertyRepository:
        return self._repository[tested.__qualname__]

    def get_results(self, tested, summary=True):
        name = tested.__qualname__
        if name not in self._repository:
            raise KeyError(f"{name} not in tested transformers {self._repository.keys()}")
        if summary:
            return self.get_repository(tested).positive_properties
        return self.get_repository(tested).results

    def add_properties_to_function(self, tested):
        test_repo = self._repository[tested.__qualname__]

        def wrapped(arg):
            result = SENTINEL

            for test in test_repo.get_properties_to_test():
                if test.__qualname__ not in self.get_repository(tested).results:
                    if len(test_repo._results) > 0:
                        with set_loggers_to_error_only():
                            result, *payload = test(tested, arg, result=result)
                    else:
                        result, *payload = test(tested, arg, result=result)
                    if len(payload) == 2:
                        has_prop, notes = payload
                    else:
                        has_prop = payload[0]
                        notes = None
                    test_repo.add_result(
                        PropertyTestResult(
                            prop=test,
                            has_property=has_prop,
                            tested_callable_name=tested.__qualname__,
                            notes=notes,
                        )
                    )
            if result is SENTINEL:
                return tested(arg)
            else:
                return result

        return wrapped

    @classmethod
    def create_for_properties(cls, properties: Collection[PropertyTest]):
        return cls(
            defaultdict(lambda: VerbosePropertyRepository.create_for_properties(properties))
        )


def depends(*on):
    if len(on) == 0:
        raise ValueError("@depends(on=[..]) or @depends(on=prop) is the required call")

    def inner(f):
        setattr(f, "depends_on", getattr(f, "depends_on", set()) | set(on))
        return f

    return inner



from uhura.base import Readable, Writable


def test_sync_read():
    class Reader(Readable[int]):
        async def read(self) -> int:
            return 1

    assert Reader().sync_read() == 1


def test_sync_write():
    has_been_written = False

    class Writer(Writable[int]):
        async def write(self, obj):
            nonlocal has_been_written
            has_been_written = True
            pass

    Writer().sync_write(1)
    assert has_been_written



from inspect import (
    isasyncgenfunction,
    iscoroutinefunction,
    isgeneratorfunction,
    isgenerator,
    isawaitable,
    isasyncgen,
)
from functools import wraps
from typing import Callable, TypeVar


_A = TypeVar("_A")
_B = TypeVar("_B")
_C = TypeVar("_C")


def compose(before: Callable[[_A], _B], after: Callable[[_B], _C]) -> Callable[[_A], _C]:
    if iscoroutinefunction(before):
        return _awaitable_compose(before, after)
    if isasyncgenfunction(before):
        return _asyncgen_compose(before, after)
    if isgeneratorfunction(before):
        return _generator_compose(before, after)
    return _compose(before, after)


def _awaitable_compose(before, after):
    @wraps(before)
    async def composed(*args, **kwargs):
        return after(await before(*args, **kwargs))

    return composed


def _asyncgen_compose(before, after):
    @wraps(before)
    async def composed(*args, **kwargs):
        async for x in before(*args, **kwargs):
            yield after(x)

    return composed


def _generator_compose(before, after):
    @wraps(before)
    def composed(*args, **kwargs):
        for x in before(*args, **kwargs):
            yield after(x)

    return composed


def _compose(before, after):
    @wraps(before)
    def composed(*args, **kwargs):
        return after(before(*args, **kwargs))

    return composed


def async_unit(obj):
    if isgenerator(obj):
        return _asyncgen_unit(obj)
    if isawaitable(obj) or isasyncgen(obj):
        return obj
    if callable(obj):
        return _function_to_coro(obj)
    return _awaitable_unit(obj)


def _awaitable_unit(value):
    async def unit():
        return value

    return unit()


def _asyncgen_unit(gen):
    async def unit():
        for x in gen:
            yield x

    return unit()


def _function_to_coro(f):
    async def inner(*args, **kwargs):
        return f(*args, **kwargs)

    return inner


def match_function_type(wrapped):
    def decorator(f):
        if iscoroutinefunction(wrapped):

            async def inner(*args, **kwargs):
                return await f(*args, **kwargs)

        elif isgeneratorfunction(wrapped):

            def inner(*args, **kwargs):
                yield from f(*args, **kwargs)

        elif isasyncgenfunction(wrapped):

            async def inner(*args, **kwargs):
                async for item in f(*args, **kwargs):
                    yield item

        else:
            inner = f
        return inner

    return decorator



import pandas as pd
from typing import Callable, Optional, Any, Tuple
from functools import wraps, partial
from uhura.properties.property_tester import AbstractPropertyTester
import inspect


def _confirm_evaluated_arg_name(signature, evaluated_arg_name: Optional[str]):
    if evaluated_arg_name is None:
        assert (
            len(signature.parameters) == 1
        ), f"No input_arg specified but the function takes more than one argument: [{signature.parameters}]"
        return list(signature.parameters)[0]
    assert (
        evaluated_arg_name in signature.parameters
    ), f"The input_arg specified [{evaluated_arg_name}] does not belong to the function signature: [{signature.parameters}]"
    return evaluated_arg_name


class _SingleArgFunctionFactory:
    def __init__(self, f: Callable, evaluated_arg_name: str):
        self._callable = f
        self._signature = inspect.Signature.from_callable(f)
        self._evaluated_arg_name = _confirm_evaluated_arg_name(self._signature, evaluated_arg_name)

    def bind_arguments(self, *args, **kwargs) -> Tuple[Any, Callable]:
        bound = self._signature.bind(*args, **kwargs)
        argument = bound.arguments[self._evaluated_arg_name]

        @wraps(self._callable)
        def inner(arg=argument):
            bound.arguments[self._evaluated_arg_name] = arg
            return self._callable(*bound.args, **bound.kwargs)

        return argument, inner


class Transformer:
    tester: Optional[AbstractPropertyTester] = None

    def __init__(self, function: Callable, evaluated_arg_name: Optional[str] = None):
        self._function = function
        self._single_arg_factory = _SingleArgFunctionFactory(function, evaluated_arg_name)

    def __call__(self, *args: Any, **kwargs: Any) -> pd.DataFrame:
        if self.tester is None:
            return self._function(*args, **kwargs)
        else:
            arg, single_arg_function = self._single_arg_factory.bind_arguments(*args, **kwargs)
            return self.tester.add_properties_to_function(single_arg_function)(arg)


def transformer(function: Optional[Callable] = None, evaluated_arg_name: Optional[str] = None):
    """
    A decorator to mark transform functions as part of a data pipeline. These functions can then be assessed
    programmatically to identify what properties they have.

    For single argument functions you can decorate with just '@transformer'

    >>> @transformer
        def add_one(value: int) -> int:
            return value + 1

    For multiple argument functions you will need to specify the 'main' argument, which is the dataset being transformed.

    >>> @transformer(evaluated_arg_name="text")
        def remove_words(text: str, words_to_remove: list) -> str:
            ...

    When not being tested, these decorators act the same as the underlying function.
    """
    if function is None:
        return partial(transformer, evaluated_arg_name=evaluated_arg_name)
    return wraps(function)(Transformer(function=function, evaluated_arg_name=evaluated_arg_name))



import asyncio

import pytest

from uhura import Readable, Writable
from functools import wraps


class ReadOne(Readable[int]):
    async def read(self):
        return 1


def _make_read_two(cls):
    @wraps(cls.read)
    async def read(self):
        return 2

    cls.read = read
    pass


def test_can_add_and_remove_decorators_to_existing_readables():
    read_one = ReadOne()
    assert asyncio.run(read_one.read()) == 1
    Readable.add_decorator(_make_read_two)
    assert asyncio.run(read_one.read()) == 2, "Decorator failed to apply to existing readable"
    Readable.remove_last_decorator()
    assert asyncio.run(read_one.read()) == 1, "Decorator was not removed from existing readable"


def test_decorators_only_apply_one_mro_layer_deep():
    class Read1(ReadOne):
        pass

    class ReadUno(Read1):
        async def read(self):
            return 3

    with Readable.decorated_with(_make_read_two):
        assert asyncio.run(ReadUno().read()) == 3  # Not covered by uhura
        assert asyncio.run(Read1().read()) == 2
    assert asyncio.run(Read1().read()) == 1


def test_can_add_and_remove_decorators_to_new_readables():
    Readable.add_decorator(_make_read_two)

    class NewReadOne(Readable[int]):
        async def read(self):
            return 1

    new_read_one = NewReadOne()
    assert asyncio.run(new_read_one.read()) == 2, "Decorator failed to apply to new readable"
    Readable.remove_last_decorator()
    assert asyncio.run(new_read_one.read()) == 1, "Decorator was not removed from new readable"

    class OtherNewReadOne(Readable[int]):
        async def read(self):
            return 1

    other_new_read_one = OtherNewReadOne()
    assert (
        asyncio.run(other_new_read_one.read()) == 1
    ), "Decorator is still being applied to future readables!"


def _make_read_three(cls):
    @wraps(cls.read)
    async def read(self):
        return 3

    cls.read = read
    pass


def test_can_decorate_new_readables_multiple_times():
    Readable.add_decorator(_make_read_two)
    Readable.add_decorator(_make_read_three)

    class NewReadOne(Readable[int]):
        async def read(self):
            return 1

    new_read_one = NewReadOne()
    assert asyncio.run(new_read_one.read()) == 3, "Decorator failed to apply to new readable"
    Readable.remove_last_decorator()
    assert (
        asyncio.run(new_read_one.read()) == 2
    ), "Last decorator was not removed from new readable"
    Readable.remove_last_decorator()
    assert (
        asyncio.run(new_read_one.read()) == 1
    ), "All decorators were not removed from new readable"

    class OtherNewReadOne(Readable[int]):
        async def read(self):
            return 1

    other_new_read_one = OtherNewReadOne()
    assert (
        asyncio.run(other_new_read_one.read()) == 1
    ), "Decorator is still being applied to future readables!"


def test_can_decorate_existing_readables_multiple_times():
    read_one = ReadOne()
    assert asyncio.run(read_one.read()) == 1
    with Readable.decorated_with(_make_read_two):
        assert asyncio.run(read_one.read()) == 2, "Decorator failed to apply to existing readable"
        with Readable.decorated_with(_make_read_three):
            assert (
                asyncio.run(read_one.read()) == 3
            ), "Second Decorator failed to apply to existing readable"
        assert (
            asyncio.run(read_one.read()) == 2
        ), "Second Decorator was not removed from existing readable"
    assert (
        asyncio.run(read_one.read()) == 1
    ), "First Decorator was not removed from existing readable"


class DummyClient(Readable[int], Writable[int]):
    async def read(self) -> int:
        return 1

    async def write(self, obj: int):
        pass


def test_multiple_inheritance_chains_dont_mess_with_each_other():
    dummy_client = DummyClient()

    def break_write(cls):
        @wraps(cls.write)
        async def write(self, obj):
            raise ValueError

        cls.write = write
        return cls

    Readable.add_decorator(_make_read_two)
    Writable.add_decorator(break_write)
    assert asyncio.run(dummy_client.read()) == 2, "Decoration failed"
    with pytest.raises(ValueError):
        asyncio.run(dummy_client.write(1))
    Readable.remove_last_decorator()
    with pytest.raises(ValueError):
        asyncio.run(dummy_client.write(2))
    Writable.remove_last_decorator()


def test_test_multiple_inheritance_future_works():
    Readable.add_decorator(_make_read_two)

    def break_write(cls):
        @wraps(cls.write)
        async def write(self, obj):
            raise ValueError

        cls.write = write
        return cls

    Writable.add_decorator(break_write)

    class NewReadable(Readable):
        async def read(self):
            pass

    Readable.remove_last_decorator()
    Writable.remove_last_decorator()



import pytest
from uhura.pandas_tools import compare_data
import pandas as pd

_basic_series = pd.Series(data=list(map(float, range(5))), index=list(reversed(list(range(5)))))


def test_series_len_check():
    with pytest.raises(AssertionError):
        compare_data(_basic_series.iloc[:4], _basic_series)


def test_series_index_check():
    with pytest.raises(AssertionError):
        compare_data(_basic_series.reset_index(), _basic_series)


def test_series_dtype_check():
    with pytest.raises(AssertionError):
        compare_data(_basic_series.astype(int), _basic_series)


def test_series_hash_check():
    _other_series = pd.Series(
        data=list(map(float, range(5))), index=list(reversed(list(range(5))))
    )
    _other_series[2] = 0.45
    with pytest.raises(AssertionError):
        compare_data(_basic_series, _other_series)


def test_skips_unhashable_series():
    unhashable_series = pd.Series(data=[list(range(4)) for _ in range(5)])
    compare_data(unhashable_series, unhashable_series)


_basic_df = pd.DataFrame(
    data=[(i, float(i), str(i)) for i in range(5)],
    columns=["integer", "float", "string"],
    index=list(reversed(list(range(5)))),
)


def test_df_different_number_of_columns_check():
    with pytest.raises(AssertionError):
        compare_data(_basic_df, _basic_df.drop(columns="integer"))


def test_df_different_length_check():
    with pytest.raises(AssertionError):
        compare_data(_basic_df, _basic_df.iloc[:4])


def test_df_different_dtype():
    with pytest.raises(AssertionError):
        compare_data(_basic_df, _basic_df.astype(int))


def test_df_different_hash():
    _different_df = pd.DataFrame(
        data=[(i, float(i), str(i)) for i in range(5)],
        columns=["integer", "float", "string"],
        index=list(reversed(list(range(5)))),
    )
    _different_df.iloc[0, 1] = 45
    with pytest.raises(AssertionError):
        compare_data(_different_df, _basic_df)



from functools import partial, wraps
from uhura.base import Readable, Writable
from uhura.composition import match_function_type
from inspect import Signature
from typing import Optional


def uhura_reader(function):
    """
    Decorator to convert functions into uhura readables. This will not change the type signature of the function.
    The cache key will default to the function's qualname.
    example:

    >>> @uhura_reader
    ... def read_int():
    ...     return 1
    """
    signature = Signature.from_callable(function)

    class _SingleFunctionReadable(Readable):
        def __init__(self, *args, **kwargs):
            self.params = signature.bind(*args, **kwargs)

        @match_function_type(function)
        def read(self):
            return function(*self.params.args, **self.params.kwargs)

        def cache_key(self):
            return function.__qualname__ + ".pkl"

    @wraps(function)
    @match_function_type(function)
    def wrapped(*args, **kwargs):
        return _SingleFunctionReadable(*args, **kwargs).read()

    return wrapped


def uhura_writer(function=None, output_arg: Optional[str] = None):
    """
    Decorator to convert functions into uhura writers. This will not change the type signature of the function.
    The cache key will default to the function's qualname. If the function only takes a single argument
    it will be assumed that this argument is what is being written out. Otherwise you will need to specify an
    output_arg (as a string) for the argument of interest.

    examples:

    >>> @uhura_writer
    ... def write_int(integer):
    ...     pass

    >>> @uhura_writer("integer")
    ... def write_integer_and_log_it(integer, logger):
    ...     pass

    """
    if function is None:
        return partial(uhura_writer, output_arg=output_arg)
    signature = Signature.from_callable(function)
    if len(signature.parameters) > 1 and output_arg not in signature.parameters:
        raise ValueError(f"Invalid/missing output_arg for writer function {function}")

    class _SingleFunctionWritable(Writable):
        def __init__(self, params):
            self.params = params

        @match_function_type(function)
        def write(self, obj):
            self.params.arguments[output_arg] = obj
            return function(*self.params.args, **self.params.kwargs)

        def cache_key(self):
            return function.__qualname__ + ".pkl"

    @wraps(function)
    @match_function_type(function)
    def wrapped(*args, **kwargs):
        bound = signature.bind(*args, **kwargs)
        if output_arg:
            written_arg = bound.arguments[output_arg]
            return _SingleFunctionWritable(bound).write(written_arg)
        return _SingleFunctionWritable(bound).write(*bound.args, **bound.kwargs)

    return wrapped



import pandas as pd
import numpy as np
from uhura.properties.homomorphic_hash import homomorphic_hash
import pytest


@pytest.fixture
def example_messy_df():
    return pd.DataFrame(index=np.arange(10), data={"a": [np.random.random(10) for _ in range(10)]})


def test_can_hash_unhashable(example_messy_df):
    assert homomorphic_hash(example_messy_df)


def test_hash_is_homomorphic(example_messy_df):
    part1 = homomorphic_hash(example_messy_df.loc[:4])
    part2 = homomorphic_hash(example_messy_df.loc[5:])
    assert part1 ^ part2 == homomorphic_hash(example_messy_df)


def test_hash_is_order_invariant(example_messy_df):
    assert homomorphic_hash(example_messy_df) == homomorphic_hash(example_messy_df.sample(frac=1))



from typing import Callable, List, Optional
import pandas as pd
import numpy as np

from uhura.properties.homomorphic_hash import homomorphic_hash
from uhura.properties.property_tester import depends, SENTINEL


def _get_batched_hash(tested, arg, number_of_batches=5):
    batch_size = max(arg.shape[0] // number_of_batches, 1)
    batch_idxs = list(range(0, arg.shape[0], batch_size))
    batch_idxs.append(batch_idxs[-1] + batch_size)
    # we are doing everything on a single line to avoid doubling the memory needs for `arg`
    return np.bitwise_xor.reduce(
        np.array(
            [
                homomorphic_hash(tested(arg.iloc[batch_idxs[idx] : batch_idxs[idx + 1]]))
                for idx in range(len(batch_idxs[:-1]))
            ]
        )
    )


def create_custom_debugger(
    notes_creator: Optional[Callable[[pd.DataFrame, Optional[pd.DataFrame]], str]]
):
    def debugger(tested, arg: pd.DataFrame, result: Optional[pd.DataFrame] = None):
        notes_arg = notes_creator(arg, None) if notes_creator is not None else ""
        notes = " - ".join([f" pre [{tested.__name__}]:", notes_arg])
        result = tested(arg.copy(deep=True)) if result is SENTINEL else result
        notes_result = notes_creator(arg, result) if notes_creator is not None else ""
        notes += " - ".join([f" post [{tested.__name__}]:", notes_result])
        return result, True, notes

    return debugger


def pure(tested, arg: pd.DataFrame, result: Optional[pd.DataFrame] = None):
    """Test whether a function modifies the dataframe it runs over (ie `arg`)

    Args:
        tested (Callable): the function for which we are testing the property
        arg (pd.DataFrame): the dataframe over which we want to run the function being tested
        result (pd.DataFrame): the result of `tested(arg)` this may already be passed as an argument in case it is
            available and we don't need to run again `tested(arg)` to check the property. So, that we can avoid
            running `tested` unnecessarily.
    Returns:
        pd.DataFrame: The output of running the tested function on the main dataframe ie `tested(arg)`
        bool: It tells if the property is respected
    """
    initial_hash = homomorphic_hash(arg)
    result = tested(arg)
    return result, initial_hash == homomorphic_hash(arg)


@depends(pure)
def homomorphic(tested, arg: pd.DataFrame, result: Optional[pd.DataFrame] = None):
    result = tested(arg) if result is SENTINEL else result
    batch_hash = _get_batched_hash(tested, arg)
    property_respected = homomorphic_hash(result) == batch_hash
    return result, property_respected


@depends(pure)
def idempotent(tested, arg: pd.DataFrame, result: Optional[pd.DataFrame] = None):
    result = tested(arg) if result is SENTINEL else result
    result_hash = homomorphic_hash(result)
    return result, result_hash == homomorphic_hash(tested(result))


def order_invariant(tested, arg: pd.DataFrame, result: Optional[pd.DataFrame] = None):
    """Check whether `tested` output is affected by the order of the rows in `arg`"""
    # `tested` can modify `arg` inplace so to avoid copying it we have to run it against
    # the shuffled version of the `arg` dataframe before we run it against the actual
    # dataframe
    shuffled_result = homomorphic_hash(tested(arg.sample(frac=1)).sort_index())
    result = tested(arg) if result is SENTINEL else result
    return result, homomorphic_hash(result.sort_index()) == shuffled_result


def adds_columns(tested, arg: pd.DataFrame, result: Optional[pd.DataFrame] = None):
    input_columns = set(arg.columns)
    result = result if isinstance(result, pd.DataFrame) else tested(arg)
    result_columns = set(result.columns)
    return result, bool(result_columns - input_columns)


def drops_columns(tested, arg: pd.DataFrame, result: Optional[pd.DataFrame] = None):
    input_columns = set(arg.columns)
    result = tested(arg) if result is SENTINEL else result
    result_columns = set(result.columns)
    return result, bool(input_columns - result_columns)


def preserves_row_count(tested, arg: pd.DataFrame, result: Optional[pd.DataFrame] = None):
    start_len = arg.shape[0]
    result = tested(arg) if result is SENTINEL else result
    return result, start_len == result.shape[0]


def preserves_index_identity(tested, arg: pd.DataFrame, result: Optional[pd.DataFrame] = None):
    arg_index = arg.index
    result = tested(arg) if result is SENTINEL else result
    return result, result.index.equals(arg_index)


def preserves_types(tested, arg: pd.DataFrame, result: Optional[pd.DataFrame] = None):
    start_types = set(arg.columns)
    result = tested(arg) if result is SENTINEL else result
    end_types = set(result.columns)
    intersection = start_types & end_types
    return result, (arg[intersection].dtypes == result[intersection].dtypes).all()


def preserves_index_types(tested, arg: pd.DataFrame, result: Optional[pd.DataFrame] = None):
    start_index = set(arg.index.names)
    result = tested(arg) if result is SENTINEL else result
    end_index = set(result.index.names)
    intersection = start_index & end_index
    intersection = list(set(arg.index.names).intersection(result.index.names))
    respected = type(arg.index) is type(result.index)  # noqa: E721
    if respected:
        if isinstance(arg.index, pd.MultiIndex):
            respected = (
                arg.index.dtypes.loc[intersection] == result.index.dtypes.loc[intersection]
            ).all()
        else:
            respected = (len(intersection) > 0) & (arg.index.dtype == result.index.dtype)
    return result, respected


def get_columns_not_modified_property(columns: List[str]):
    @depends(order_invariant)
    def columns_not_modified(tested, arg: pd.DataFrame, result: Optional[pd.DataFrame] = None):
        arg_sorted = arg.sort_index()
        original_values = {
            c: homomorphic_hash(arg_sorted[c]) for c in columns if c in arg_sorted.columns
        }
        result = tested(arg_sorted) if result is SENTINEL else result
        result_sorted = result.sort_index()
        # TODO we should check the two below as property dependencies here. However,
        # testing multiple properties requires purity, or a refactor. Purity limits the scope
        # of properties and a refactor requires time
        if result_sorted.shape[0] != arg.shape[0]:
            # TODO we should use `preserves_row_count` here
            return result, False, "Cannot test, the dataframe length has changed"
        if not result_sorted.index.equals(arg_sorted.index):
            # TODO we should use `preserves_index_identity` here
            return result, False, "Cannot test, the dataframe index has been changed"
        new_values = {
            c: homomorphic_hash(result_sorted[c])
            for c in original_values
            if c in result_sorted.columns
        }
        changes = []
        for c, vals in original_values.items():
            if c not in new_values:
                changes.append(f"{c} dropped")
            else:
                if not new_values[c] == vals:
                    changes.append(f"{c} modified")

        return result, len(changes) == 0, ", ".join(changes)

    return columns_not_modified


DEFAULT_PROPERTIES = tuple(
    [
        pure,
        homomorphic,
        idempotent,
        order_invariant,
        adds_columns,
        drops_columns,
        preserves_row_count,
        preserves_types,
        preserves_index_types,
    ]
)



from collections import defaultdict
from typing import Any

import pandas as pd

from uhura.comparison import Comparer


def compare_data(data1: Any, data2: Any):
    assert type(data1) is type(data2), f"Data types do not match: {type(data1)} vs {type(data2)}"
    datatype = type(data1)
    if datatype not in COMPARISON_LOOKUP:
        assert data1 == data2, "Observed != Expected"
    else:
        COMPARISON_LOOKUP[datatype](data1, data2)


class PandasComparer(Comparer):
    def base_compare(self, actual, expected):
        return compare_data(actual, expected)


COMPARISON_LOOKUP = {
    pd.DataFrame: pd.testing.assert_frame_equal,
    pd.Series: pd.testing.assert_series_equal,
}

pandas_comparator = defaultdict(PandasComparer)
